#include <iostream>
#include <fstream>
#include <string>

#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/exception.hpp>

#include "otrdownloader.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace libtorrent;
using namespace std;

#define SETTINGS_GROUP "settings"
#define CONFIGURATION_FILE ".otrdownloader"

string to_string(float v, int width = 0, int precision = 3, bool zerofill = false)
{
  stringstream s;
  s.precision(precision);
  //s.flags(ios_base::right);
  if (width != 0)
    s.width(width);
  if (zerofill)
    s.fill('0');
  s << v;
  return s.str();
}

string add_suffix(float val)
{
  const char* prefix[] = {"B", "kB", "MB", "GB", "TB"};
  const int num_prefix = sizeof(prefix) / sizeof(const char*);
  for (int i = 0; i < num_prefix; ++i)
    {
      if (fabs(val) < 1024.f)
	return to_string(val) + prefix[i];
      val /= 1024.f;
    }
  return to_string(val) + "PB";
}

string add_timesuffix(float val)
{
  const char* prefix[] = {"s", "min"};
  const int num_prefix = sizeof(prefix) / sizeof(const char*);
  string result = "";
  for (int i = 0; i < num_prefix; ++i)
    {
      if (fabs(val) < 60.f)
	return to_string(floor(val)) + result + prefix[i];
      result = ":" + to_string(fmod(floor(val), 60.f), 2, 2, true);
      val /= 60.f;
    }
  return to_string(floor(val)) + result + "h";
}

OTRDownloader::OTRDownloader()
  : _pWindow(NULL),
    _pTreeViewTorrents(NULL),
    _pTreeViewPeers(NULL)
{
  //Load the Glade file and instiate its widgets:
  try
  {
    _refXml = Gnome::Glade::Xml::create(DATA_DIR "/otrdownloader.glade");
  }
  catch(const Gnome::Glade::XmlError& ex)
  {
    cerr << ex.what() << endl;
    exit(1);
  }

  //namespace fs = boost::filesystem;
  //fs::path::default_name_check(fs::no_check);

  _refXml->get_widget("window1", _pWindow);

  _refXml->connect_clicked("oeffnen1", sigc::mem_fun(*this, &OTRDownloader::on_oeffnen1_activate));
  _refXml->connect_clicked("einstellungen1", sigc::mem_fun(*this, &OTRDownloader::on_einstellungen1_activate));
  _refXml->connect_clicked("info1", sigc::mem_fun(*this, &OTRDownloader::on_info1_activate));

  Gtk::ToolButton* pButton = NULL;
  _refXml->get_widget("open_toolbutton", pButton);
  if(pButton)
    pButton->signal_clicked().connect(sigc::mem_fun(*this, &OTRDownloader::on_oeffnen1_activate));
  pButton = NULL;
  _refXml->get_widget("delete_toolbutton", pButton);
  if(pButton)
    pButton->signal_clicked().connect(sigc::mem_fun(*this, &OTRDownloader::on_delete_activate));
  pButton = NULL;
  _refXml->get_widget("resume_toolbutton", pButton);
  if(pButton)
    pButton->signal_clicked().connect(sigc::mem_fun(*this, &OTRDownloader::on_resume_activate));
  pButton = NULL;
  _refXml->get_widget("pause_toolbutton", pButton);
  if(pButton)
    pButton->signal_clicked().connect(sigc::mem_fun(*this, &OTRDownloader::on_pause_activate));

  _refXml->connect_clicked("beenden1", sigc::mem_fun(*this, &OTRDownloader::on_beenden1_activate));
  _pWindow->signal_delete_event().connect(sigc::mem_fun(*this, &OTRDownloader::on_delete_event));

  _refXml->get_widget("treeview1", _pTreeViewTorrents);
  if (_pTreeViewTorrents) {
    _refListStoreTorrents = Gtk::ListStore::create(_modelTorrents);
    _pTreeViewTorrents->set_model(_refListStoreTorrents);
    _pTreeViewTorrents->append_column("Dateiname", _modelTorrents.filename);
    _pTreeViewTorrents->append_column("Downloadrate", _modelTorrents.downloadrate);
    _pTreeViewTorrents->append_column("Download", _modelTorrents.totaldownload);
    _pTreeViewTorrents->append_column("restliche Zeit", _modelTorrents.remainingtime);
    _pTreeViewTorrents->append_column("Uploadrate", _modelTorrents.uploadrate);
    Gtk::CellRendererProgress* pRenderer = new Gtk::CellRendererProgress;
    int cols_count = _pTreeViewTorrents->append_column("Progress", *pRenderer);
    Gtk::TreeViewColumn* pColumn = _pTreeViewTorrents->get_column(cols_count - 1);
    if(pColumn)
      {
#ifdef GLIBMM_PROPERTIES_ENABLED
	pColumn->add_attribute(pRenderer->property_value(), _modelTorrents.progress);
#else
	pColumn->add_attribute(*pRenderer, "value", _modelTorrents.progress);
#endif
      }
    for(guint i = 0; i < cols_count; i++)
      {
	Gtk::TreeView::Column* pColumn = _pTreeViewTorrents->get_column(i);
	
	pColumn->set_sort_indicator(true);
	pColumn->set_reorderable(true);
	pColumn->set_resizable(true);
	pColumn->set_sort_column_id(i);
      }
    _pTreeViewTorrents->set_headers_clickable(true);
    _refListStoreTorrents->set_sort_column_id(0, Gtk::SORT_ASCENDING);
  }
   
  _refXml->get_widget("treeview2", _pTreeViewPeers);
  if (_pTreeViewPeers) {
    _refListStorePeers = Gtk::ListStore::create(_modelPeers);
    _pTreeViewPeers->set_model(_refListStorePeers);
    _pTreeViewPeers->append_column("IP", _modelPeers.ip);
    _pTreeViewPeers->append_column("Downloadrate", _modelPeers.downloadrate);
    _pTreeViewPeers->append_column("Download", _modelPeers.totaldownload);
    _pTreeViewPeers->append_column("Uploadrate", _modelPeers.uploadrate);
    _pTreeViewPeers->append_column("Upload", _modelPeers.totalupload);
    _pTreeViewPeers->append_column("Status", _modelPeers.status);
    Gtk::CellRendererProgress* pRenderer = new Gtk::CellRendererProgress;
    int cols_count = _pTreeViewPeers->append_column("Progress", *pRenderer);
    Gtk::TreeViewColumn* pColumn = _pTreeViewPeers->get_column(cols_count - 1);
    if(pColumn)
      {
#ifdef GLIBMM_PROPERTIES_ENABLED
	pColumn->add_attribute(pRenderer->property_value(), _modelPeers.progress);
#else
	pColumn->add_attribute(*pRenderer, "value", _modelPeers.progress);
#endif
      }
    
    for(guint i = 0; i < cols_count; i++)
      {
	Gtk::TreeView::Column* pColumn = _pTreeViewPeers->get_column(i);
	
	pColumn->set_sort_indicator(true);
	pColumn->set_reorderable(true);
	pColumn->set_resizable(true);
	pColumn->set_sort_column_id(i);
      }
    _pTreeViewPeers->set_headers_clickable(true);
    _refListStorePeers->set_sort_column_id(0, Gtk::SORT_ASCENDING);
  }
  
  load_preferences();

  Glib::signal_timeout().connect( sigc::mem_fun(*this, &OTRDownloader::on_timeout), 1000);
}

OTRDownloader::~OTRDownloader()
{
}

void OTRDownloader::addTorrent(const string& filename)
{
  using namespace libtorrent;
  boost::filesystem::path save_path("./");

  ifstream in(filename.c_str(), ios_base::binary);
  in.unsetf(ios_base::skipws);
  entry e = bdecode(istream_iterator<char>(in), istream_iterator<char>());
  
  torrent_info t(e);

  entry resume_data;
  try
    {
      stringstream s;
      s << t.name() << ".fastresume";
      boost::filesystem::ifstream resume_file(save_path / s.str(),
					      ios_base::binary);
      resume_file.unsetf(ios_base::skipws);
      resume_data = bdecode(istream_iterator<char>(resume_file),
			    istream_iterator<char>());
    }
  catch (invalid_encoding&) {}
  catch (boost::filesystem::filesystem_error&) {}
  
  torrent_handle h = _session.add_torrent(t, save_path, resume_data);
  torrent_status s = h.status();

  Gtk::TreeModel::Row row = *(_refListStoreTorrents->append());
  row[_modelTorrents.filename] = h.get_torrent_info().name();
  string downloadrate = (add_suffix(s.download_rate)+"/s");
  row[_modelTorrents.downloadrate] = downloadrate;
  string totaldownload = (add_suffix(s.total_wanted_done)+"/"+add_suffix(s.total_wanted));
  row[_modelTorrents.totaldownload] = totaldownload;
  string remainingtime;
  if (s.download_rate > 0)
    remainingtime = add_timesuffix((s.total_wanted - s.total_wanted_done) / s.download_rate);
  else
    remainingtime = "oo";
  row[_modelTorrents.remainingtime] = remainingtime;
  string uploadrate = (add_suffix(s.upload_rate)+"/s");
  row[_modelTorrents.uploadrate] = uploadrate;
  row[_modelTorrents.progress] = (int)(s.progress*100.);
  row[_modelTorrents.handle] = h;
}

void OTRDownloader::on_oeffnen1_activate()
{
  Gtk::FileChooserDialog dialog("Torrent-Datei ffnen",
				Gtk::FILE_CHOOSER_ACTION_OPEN);
  dialog.set_transient_for(*_pWindow);

  //Add response buttons the the dialog:
  dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

  //Add filters, so that only certain file types can be selected:

  Gtk::FileFilter filter_torrent;
  filter_torrent.set_name("Torrent Dateien");
  //  filter_torrent.add_pattern("*.torrent");
  filter_torrent.add_mime_type("application/x-bittorrent");
  dialog.add_filter(filter_torrent);

  Gtk::FileFilter filter_any;
  filter_any.set_name("Any files");
  filter_any.add_pattern("*");
  dialog.add_filter(filter_any);

  //Show the dialog and wait for a user response:
  int result = dialog.run();

  //Handle the response:
  switch(result)
  {
    case(Gtk::RESPONSE_OK):
    {
      addTorrent(dialog.get_filename());
      break;
    }
    default:
    {
      break;
    }
  }
}

void OTRDownloader::load_preferences()
{
  boost::filesystem::path home_path(getenv("HOME"));

  try {
    _configFile.load_from_file((home_path / CONFIGURATION_FILE).string(), Glib::KEY_FILE_KEEP_COMMENTS);
  } catch (Glib::FileError& e) {
    cerr << "ERROR:" << e.what() << endl;
    cerr << "Loading default values!" << endl;
  } catch (Glib::KeyFileError& e) {
    cerr << "ERROR:" << e.what() << endl;
  }
  Gtk::SpinButton* pButton = NULL;
  _refXml->get_widget("maxdownloadrate_spinbutton", pButton);
  if (pButton) {
    int maxdownloadrateKB;
    try {
      maxdownloadrateKB = _configFile.get_integer(SETTINGS_GROUP, "maxdownloadrate");
    } catch (Glib::KeyFileError& e) {
      maxdownloadrateKB = 0;
      _configFile.set_integer(SETTINGS_GROUP, "maxdownloadrate", maxdownloadrateKB);
      _configFile.set_comment(SETTINGS_GROUP, "maxdownloadrate", "Maximum download rate in KB.");
    }
    pButton->set_value(maxdownloadrateKB);
    int value = maxdownloadrateKB * 1024;
    _session.set_download_rate_limit((value <= 0)?-1:value);
  }
  _refXml->get_widget("maxuploadrate_spinbutton", pButton);
  if (pButton) {
    int maxuploadrateKB;
    try {
      maxuploadrateKB = _configFile.get_integer(SETTINGS_GROUP, "maxuploadrate");
    } catch (Glib::KeyFileError& e) {
      maxuploadrateKB = 0;
      _configFile.set_integer(SETTINGS_GROUP, "maxuploadrate", maxuploadrateKB);
      _configFile.set_comment(SETTINGS_GROUP, "maxuploadrate", "Maximum upload rate in KB.");
    }
    pButton->set_value(maxuploadrateKB);
    int value = maxuploadrateKB * 1024;
    _session.set_upload_rate_limit((value <= 0)?-1:value);
  }
  _refXml->get_widget("maxuploadslots_spinbutton", pButton);
  if (pButton) {
    int maxuploadslots;
    try {
      maxuploadslots = _configFile.get_integer(SETTINGS_GROUP, "maxuploadslots");
    } catch (Glib::KeyFileError& e) {
      maxuploadslots = 1;
      _configFile.set_integer(SETTINGS_GROUP, "maxuploadslots", maxuploadslots);
      _configFile.set_comment(SETTINGS_GROUP, "maxuploadslots", "Maximum number of upload slots.");
    }
    pButton->set_value(maxuploadslots);
    _session.set_max_uploads((maxuploadslots <= 0)?-1:maxuploadslots);
  }
}

void OTRDownloader::save_preferences()
{
  Gtk::SpinButton* pButton = NULL;
  _refXml->get_widget("maxdownloadrate_spinbutton", pButton);
  if (pButton) {
    int maxdownloadrateKB = (int)pButton->get_value();
    _configFile.set_integer(SETTINGS_GROUP, "maxdownloadrate", maxdownloadrateKB);
    int value = maxdownloadrateKB * 1024;
    _session.set_download_rate_limit((value <= 0)?-1:value);
  }
  _refXml->get_widget("maxuploadrate_spinbutton", pButton);
  if (pButton) {
    int maxuploadrateKB = (int)pButton->get_value();
    _configFile.set_integer(SETTINGS_GROUP, "maxuploadrate", maxuploadrateKB);
    int value = maxuploadrateKB * 1024;
    _session.set_upload_rate_limit((value <= 0)?-1:value);
  }
  _refXml->get_widget("maxuploadslots_spinbutton", pButton);
  if (pButton) {
    int maxuploadslots = (int)pButton->get_value();
    _configFile.set_integer(SETTINGS_GROUP, "maxuploadslots", maxuploadslots);
    _session.set_max_uploads((maxuploadslots <= 0)?-1:maxuploadslots);
  }
}

void OTRDownloader::on_einstellungen1_activate()
{
  Gtk::Dialog* pDialog = NULL;
  _refXml->get_widget("pref_dialog1", pDialog);
  if (pDialog) {
    int result;
    while ((result = pDialog->run()) == Gtk::RESPONSE_APPLY)
      save_preferences();
    switch(result)
      {
      case(Gtk::RESPONSE_OK):
	{
	  save_preferences();
	  break;
	}
      default:
	{
	  break;
	}
      }
    pDialog->hide();
  }
}

void OTRDownloader::on_info1_activate()
{
  Gtk::AboutDialog* pDialog = NULL;
  _refXml->get_widget("aboutdialog1", pDialog);
  if (pDialog) {
    pDialog->set_version(PACKAGE_VERSION);
    pDialog->run();
    pDialog->hide();
  }
}

void OTRDownloader::on_delete_activate()
{
  Glib::RefPtr<Gtk::TreeSelection> selection = _pTreeViewTorrents->get_selection();
  if (selection->count_selected_rows() == 1) {

    Gtk::MessageDialog dialog(*_pWindow, "",
			      false /* use_markup */, Gtk::MESSAGE_WARNING,
			      Gtk::BUTTONS_OK_CANCEL);
    dialog.set_secondary_text("Torrent wirklich aus der Liste entfernen?");
    
    int result = dialog.run();
    
    //Handle the response:
    switch(result)
      {
      case(Gtk::RESPONSE_OK):
	{
	  Gtk::TreeIter iter = selection->get_selected();
	  Gtk::TreeModel::Row row = *iter;

	  libtorrent::torrent_handle h = row[_modelTorrents.handle];

	  entry data = h.write_resume_data();
	  stringstream s;
	  s << h.get_torrent_info().name() << ".fastresume";
	  boost::filesystem::ofstream out(h.save_path() / s.str(), ios_base::binary);
	  out.unsetf(ios_base::skipws);
	  bencode(ostream_iterator<char>(out), data);

	  _session.remove_torrent(h);
	  _refListStoreTorrents->erase(iter);
	  _refListStorePeers->clear();
	  break;
	}
      default:
	{
	  break;
	}
      }
  }
}

void OTRDownloader::on_resume_activate()
{
  Glib::RefPtr<Gtk::TreeSelection> selection = _pTreeViewTorrents->get_selection();
  if (selection->count_selected_rows() == 1) {
    Gtk::TreeIter iter = selection->get_selected();
    Gtk::TreeModel::Row row = *iter;
    libtorrent::torrent_handle h = row[_modelTorrents.handle];
    h.resume();
  }
}

void OTRDownloader::on_pause_activate()
{
  Glib::RefPtr<Gtk::TreeSelection> selection = _pTreeViewTorrents->get_selection();
  if (selection->count_selected_rows() == 1) {
    Gtk::TreeIter iter = selection->get_selected();
    Gtk::TreeModel::Row row = *iter;
    libtorrent::torrent_handle h = row[_modelTorrents.handle];
    h.pause();
  }
}

void OTRDownloader::on_beenden1_activate()
{
  Gtk::TreeNodeChildren children = _refListStoreTorrents->children();
  for (Gtk::TreeIter iter = children.begin(), end(children.end()); iter != end; ++iter)
    {
      Gtk::TreeModel::Row row = *iter;
      libtorrent::torrent_handle h = row[_modelTorrents.handle];

      entry data = h.write_resume_data();
      stringstream s;
      s << h.get_torrent_info().name() << ".fastresume";
      boost::filesystem::ofstream out(h.save_path() / s.str(), ios_base::binary);
      out.unsetf(ios_base::skipws);
      bencode(ostream_iterator<char>(out), data);
      
      _session.remove_torrent(h);
    }
  boost::filesystem::path home_path(getenv("HOME"));
  boost::filesystem::fstream filestr(home_path / CONFIGURATION_FILE, ios_base::out);

  filestr << _configFile.to_data();
  filestr.close();
  Gtk::Main::quit();
}

bool OTRDownloader::on_delete_event(GdkEventAny* event)
{
  on_beenden1_activate();
  return true;
}

bool OTRDownloader::on_timeout()
{
  Gtk::TreeNodeChildren children = _refListStoreTorrents->children();
  for (Gtk::TreeIter iter = children.begin(), end(children.end()); iter != end; ++iter)
    {
      Gtk::TreeModel::Row row = *iter;
      libtorrent::torrent_handle h = row[_modelTorrents.handle];
      torrent_status s = h.status();
      string downloadrate = (add_suffix(s.download_rate)+"/s");
      row[_modelTorrents.downloadrate] = downloadrate;
      string totaldownload = (add_suffix(s.total_wanted_done)+"/"+add_suffix(s.total_wanted));
      row[_modelTorrents.totaldownload] = totaldownload;
      string remainingtime;
      if (s.download_rate > 0)
	remainingtime = add_timesuffix((s.total_wanted - s.total_wanted_done) / s.download_rate);
      else
	remainingtime = "oo";
      row[_modelTorrents.remainingtime] = remainingtime;
      string uploadrate = (add_suffix(s.upload_rate)+"/s");
      row[_modelTorrents.uploadrate] = uploadrate;
      row[_modelTorrents.progress] = (int)(s.progress * 100.); 
    }

  _refListStorePeers->clear();
  Glib::RefPtr<Gtk::TreeSelection> selection = _pTreeViewTorrents->get_selection();
  if (selection->count_selected_rows() == 1) {
    vector<peer_info> peers;
    Gtk::TreeIter iter = selection->get_selected();
    Gtk::TreeModel::Row row = *iter;
    libtorrent::torrent_handle h = row[_modelTorrents.handle];
    h.get_peer_info(peers);

    for (vector<peer_info>::const_iterator i = peers.begin(); i != peers.end(); ++i)
      {
	Gtk::TreeModel::Row row = *(_refListStorePeers->append());
	row[_modelPeers.ip] = i->ip.address().to_string();
	string totaldownload = add_suffix(i->total_download);
	row[_modelPeers.totaldownload] = totaldownload;
	string downloadrate = (i->down_speed > 0)?(add_suffix(i->down_speed)+"/s"):"";
	row[_modelPeers.downloadrate] = downloadrate;
	string totalupload = add_suffix(i->total_upload);
	row[_modelPeers.totalupload] = totalupload;
	string uploadrate = (i->up_speed > 0)?(add_suffix(i->up_speed)+"/s"):"";
	row[_modelPeers.uploadrate] = uploadrate;
	if (i->downloading_piece_index >= 0)
	  {
	    row[_modelPeers.progress] = (int)(i->downloading_progress / float(i->downloading_total) * 100.);
	  }
	else
	  {
	    row[_modelPeers.progress] = 0;
	  }
	if (i->flags & peer_info::handshake)
	  {
	    row[_modelPeers.status] = "waiting for handshake";
	  }
	else if (i->flags & peer_info::connecting)
	  {
	    row[_modelPeers.status] = "connecting to peer";
	  }
	else if (i->flags & peer_info::queued)
	  {
	    row[_modelPeers.status] = "queued";
	  }
	else
	  {
	    row[_modelPeers.status] = i->client;
	  }
      }
  }
  return true;
}

void OTRDownloader::run()
{
  _session.listen_on(make_pair(6881, 6889));
  
  if (_pWindow)
    Gtk::Main::run(*_pWindow);
}

int main (int argc, char **argv)
{
  if(!Glib::thread_supported())
    Glib::thread_init();
  Gtk::Main kit(argc, argv);
  OTRDownloader downloader;
  for (unsigned i=1; i < argc; i++)
    downloader.addTorrent(argv[i]);

  try
    {
      downloader.run();
    }
  catch (exception& e)
    {
      cerr << e.what() << "\n";
    }
  return 0;
}

