#ifndef _otrdownloader_h_
#define _otrdownloader_h_

#include <libglademm/xml.h>
#include <gtkmm.h>
#include <gtkmm/liststore.h>

#include <string>
#include <list>

#include "libtorrent/entry.hpp"
#include "libtorrent/bencode.hpp"
#include "libtorrent/session.hpp"

class OTRDownloader {
 public:
  OTRDownloader();
  ~OTRDownloader();
  void addTorrent(const std::string& filename);
  void run();
 private:
  bool on_timeout();
  void on_oeffnen1_activate();
  void on_einstellungen1_activate();
  void on_info1_activate();
  void on_delete_activate();
  void on_resume_activate();
  void on_pause_activate();
  void on_beenden1_activate();
  bool on_delete_event(GdkEventAny* event);
  void load_preferences();
  void save_preferences();

 class TorrentsModelColumns : public Gtk::TreeModel::ColumnRecord
    {
    public:
      Gtk::TreeModelColumn<Glib::ustring> filename;
      Gtk::TreeModelColumn<Glib::ustring> downloadrate;
      Gtk::TreeModelColumn<Glib::ustring> totaldownload;
      Gtk::TreeModelColumn<Glib::ustring> remainingtime;
      Gtk::TreeModelColumn<Glib::ustring> uploadrate;
      Gtk::TreeModelColumn<int> progress;
      Gtk::TreeModelColumn<libtorrent::torrent_handle> handle;
      
      
      TorrentsModelColumns() {
	add(filename);
	add(downloadrate);
	add(totaldownload);
	add(remainingtime);
	add(uploadrate);
	add(progress);
	add(handle);
      }
    };

 class PeersModelColumns : public Gtk::TreeModel::ColumnRecord
    {
    public:
      Gtk::TreeModelColumn<Glib::ustring> ip;
      Gtk::TreeModelColumn<Glib::ustring> downloadrate;
      Gtk::TreeModelColumn<Glib::ustring> totaldownload;
      Gtk::TreeModelColumn<Glib::ustring> uploadrate;
      Gtk::TreeModelColumn<Glib::ustring> totalupload;
      Gtk::TreeModelColumn<Glib::ustring> status;
      Gtk::TreeModelColumn<int> progress;
      
      PeersModelColumns() {
	add(ip);
	add(downloadrate);
	add(totaldownload);
	add(uploadrate);
	add(totalupload);
	add(status);
	add(progress);
      }
    };

  Glib::RefPtr<Gnome::Glade::Xml> _refXml;

  Gtk::Window* _pWindow;
  Gtk::TreeView* _pTreeViewTorrents;
  Gtk::TreeView* _pTreeViewPeers;
  Glib::RefPtr<Gtk::ListStore> _refListStoreTorrents;
  Glib::RefPtr<Gtk::ListStore> _refListStorePeers;

  libtorrent::session _session;

  TorrentsModelColumns _modelTorrents;
  PeersModelColumns _modelPeers;

  Glib::KeyFile _configFile;
};

#endif
